/// ===========================================================================
/// C= Parallel Programming Language Support Header
/// ===========================================================================
/// Copyright HOOPOE ProGroup, 2010-2011. All rights reserved
/// ---------------------------------------------------------------------------
///
/// GENERAL NOTES:  C= (pronounced 'see stripes') is a 
///                 Parallel C/C++ Programming Language Extension.
///                 CPX prefix stands for 'C= Parallel eXtension'
///
/// ---------------------------------------------------------------------------
/// LEGAL NOTICE: The terms and conditions of use of this file and the code 
///               herein are subject to applicable laws and an accompanying 
///               license agreement, and may vary depending on the actual 
///               purpose and form of use, including but not limited to
///               research, non-commercial, business purposes, derived 
///               works, as appropriately covered by said license agreement.
///               Customization of keywords provided in a dedicated section 
///               of this file is allowed only to resolve otherwise
///               unavoidable keyword conflicts, while all other modifications
///               are subject to explicit licensing.
/// ---------------------------------------------------------------------------
/// VERSION: 0.3
/// ---------------------------------------------------------------------------

#ifndef _CPX_HEADER_H_
#define _CPX_HEADER_H_

#include <pthread.h>
#include <semaphore.h>
#include <ucontext.h>
#include <setjmp.h>
#include <stdarg.h>
#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

/// ===========================================================================
/// C= Language Customization Part
/// ===========================================================================

/// ---------------------------------------------------------------------------
/// Custom Keywords
/// ---------------------------------------------------------------------------

/// primary operators
#define parallel                CPX_KEYWORD_PARALLEL
#define serial                  CPX_KEYWORD_SERIAL
/// language-defined functions (secondary operators)
#define wait                    CPX_KEYWORD_WAIT
#define wake                    CPX_KEYWORD_WAKE
#define pid                     CPX_KEYWORD_PID
#define pix                     CPX_KEYWORD_PIX
#define pno                     CPX_KEYWORD_PNO
/// operators to become implicit with the compiler support
#define barrier                 CPX_KEYWORD_BARRIER
/// operators to merge with C keywords with the compiler support
#define break()                 CPX_KEYWORD_BREAK()
/// try() operator is to be customized by changing its definition (below)
/// data type modifiers to become implicit with the compiler support
#define shared                  CPX_KEYWORD_SHARED

/// ---------------------------------------------------------------------------
/// Custom Macros
/// ---------------------------------------------------------------------------

/// parallel data decomposition macros
#define CPX_RANGE_START         CPX_MACRO_RANGE_START
#define CPX_RANGE_END           CPX_MACRO_RANGE_END

/// ===========================================================================
/// C= Runtime Part
/// ===========================================================================

/// C= runtime functions are prefixed with cpx_exported_
/// C= auxiliary functions are prefixed with cpx_intrinsic_
/// C= data types are prefixed with cpx_datatype_
/// C= global variables are prefixed with cpx_global_
/// C= automatic variables are prefixed with cpx_local_
/// C= enumeration constants are prefixed with cpx_const_
/// C= keyword macros are prefixed with CPX_KEYWORD_
/// C= functional macros are prefixed with CPX_MACRO_
/// C= data macros are prefixed with CPX_CONST_

/// ---------------------------------------------------------------------------
/// Internal Data Structures
/// ---------------------------------------------------------------------------

#pragma pack(push, 1)

typedef struct
{
    size_t stk_bottom;
    size_t stk_frame;

    ucontext_t regs;

} cpx_datatype_execution_context;

struct _cpx_datatype_tls_context;

typedef enum {cpx_const_task = 0, cpx_const_thread = 1} cpx_datatype_partype;

typedef struct _cpx_datatype_region_context
{
    /// backward compatibility
    size_t version;

    /// parallel execution type
    volatile cpx_datatype_partype type;

    /// operating context
    int pix;                                    /// current parallel index
    int pno;                                    /// size of parallel region
    volatile long busy;                         /// unsafe-to-return flag
    volatile long lcount;                       /// tasks to launch
    volatile long bcount;                       /// tasks on barrier

    /// context references
    struct _cpx_datatype_tls_context* head;     /// main task's TLS
    struct _cpx_datatype_tls_context* curr;     /// current TLS

    struct _cpx_datatype_region_context* head_rgn;  /// head region

    jmp_buf* idle;                                  /// idle execution context

    struct _cpx_datatype_region_context* prev_nest;  /// saved tls::rgnctx
    struct _cpx_datatype_region_context* prev_task;  /// saved task list head

    /// register context
    cpx_datatype_execution_context exectx;

} cpx_datatype_region_context;

typedef struct _cpx_datatype_tls_context
{
    /// parallel region context (in-stack)
    cpx_datatype_region_context* rgnctx;

    /// task synchronization objects
    pthread_mutex_t mutex;
    sem_t wakesig;

    /// object-free synchronization elements
    volatile struct _cpx_datatype_tls_context* prev;
    volatile struct _cpx_datatype_tls_context* next;
    
    /// asynchronous context initialization flag
    volatile int ctxset;

} cpx_datatype_tls_context;

#pragma pack(pop)

/// ---------------------------------------------------------------------------
/// Internal Runtime Data
/// ---------------------------------------------------------------------------

#define CPX_CONST_STATE_PARALLEL        0
#define CPX_CONST_STATE_SERIAL_SYNC     1
#define CPX_CONST_STATE_SERIAL_DEFAULT  2

/// ---------------------------------------------------------------------------
/// Internal Runtime Functions and Macros
/// ---------------------------------------------------------------------------

/// defines the number of function arguments to share by parallel tasks
#define CPX_MACRO_SHARED_ARGS 8

/// sets up a mode to detect returns and gotos out of parallel regions
///#define CPX_MACRO_DEBUG_MODE

/// checks if a macro was invoked with zero (false) or one argument (true)
#define CPX_MACRO_ONE_ARG(...) cpx_intrinsic_one_arg(0, ## __VA_ARGS__, 1, 0)
#define CPX_MACRO_ONE_PTR(...) cpx_intrinsic_one_ptr                         \
                                   (0, ## __VA_ARGS__, (void*)1, (void*)0)

int __attribute__((noinline)) cpx_intrinsic_one_arg
                                       (size_t _0, size_t _1, size_t _2, ...)
{
    return (int)_2;
}

int __attribute__((noinline)) cpx_intrinsic_one_ptr
                                       (size_t _0, void* _1, void* _2, ...)
{
    return (int)(size_t)_2;
}

/// fetches an optional argument
size_t __attribute__((noinline)) cpx_intrinsic_get_arg
                                          (int argc, /*size_t arg,*/ ...)
{
    va_list argptr;
    va_start(argptr, argc);

    return (size_t)va_arg(argptr, int);
}

#ifdef CPX_MACRO_DEBUG_MODE

/// obtains the current execution address
static size_t __attribute__((noinline)) cpx_intrinsic_get_instruction_pointer()
{
    return (size_t)__builtin_return_address(0);
}

#endif

cpx_datatype_tls_context* cpx_intrinsic_get_tlsctx();

/// parallel operator support functions
extern void cpx_exported_parallel(int x, cpx_datatype_region_context* rgnctx);
extern void* cpx_exported_set_exectx
                       (cpx_datatype_execution_context* exectx, void* frame);
extern int cpx_exported_pick_job();
extern int cpx_exported_pick_nested_job
                       (cpx_datatype_region_context* currctx);

/// serial operator support functions
extern void cpx_exported_sync(size_t addr);
extern void cpx_exported_unsync(size_t addr);
extern int cpx_exported_trysync(size_t addr);
extern void cpx_exported_sync_default();
extern void cpx_exported_unsync_default();

/// parallel task identification support
extern void* cpx_exported_pid();
extern int cpx_exported_pix();
extern int cpx_exported_pno();

/// parallel task synchronization support
extern void cpx_exported_wake(void* pid);
/// returns true if timed out, and false otherwise
extern int cpx_exported_wait(int argc, /*int milliseconds*/...);

/// data decomposition support
extern int cpx_exported_range_start(int size);
extern int cpx_exported_range_end(int size);

/// ===========================================================================
/// C= Language Syntax Part
/// ===========================================================================

/// ---------------------------------------------------------------------------
/// Language-Defined Operators
/// ---------------------------------------------------------------------------

#ifdef CPX_MACRO_DEBUG_MODE

/// parallel operator
#define CPX_KEYWORD_PARALLEL(...)                                            \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_PARALLEL;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_arg(1,## __VA_ARGS__, 0);   \
                                                                             \
    cpx_datatype_region_context cpx_local_rgnctx;                            \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr;                       \
                                                                             \
    struct timespec req = {0, 1000};                                         \
    struct timespec rem;                                                     \
    rem.tv_nsec = cpx_local_region_state;                                    \
                                                                             \
    cpx_local_rgnctx.exectx.stk_frame =                                      \
        (size_t)__builtin_frame_address(0) + sizeof(void*) *                 \
        (CPX_MACRO_SHARED_ARGS + 3);                                         \
                                                                             \
                                                                             \
    cpx_intrinsic_get_tlsctx()->ctxset = 0;                                  \
    getcontext(&cpx_local_rgnctx.exectx.regs);                               \
                                                                             \
    if(!cpx_intrinsic_get_tlsctx()->ctxset)                                  \
    {                                                                        \
        cpx_local_rgnctx_ptr = &cpx_local_rgnctx;                            \
        cpx_exported_parallel((int)cpx_local_opt_arg, &cpx_local_rgnctx);    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_rgnctx_ptr = cpx_intrinsic_get_tlsctx()->rgnctx;           \
                                                                             \
        ((size_t*)__builtin_frame_address(0))[CPX_MACRO_SHARED_ARGS + 2] =   \
            cpx_intrinsic_get_instruction_pointer();                         \
        ((size_t*)__builtin_frame_address(0))[1] = 0;                        \
                                                                             \
        __sync_fetch_and_sub(&cpx_local_rgnctx_ptr->head_rgn->busy, 1);      \
    }

#else

/// parallel operator
#define CPX_KEYWORD_PARALLEL(...)                                            \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_PARALLEL;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_arg(1,## __VA_ARGS__, 0);   \
                                                                             \
    cpx_datatype_region_context cpx_local_rgnctx;                            \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr;                       \
                                                                             \
    struct timespec req = {0, 1000};                                         \
    struct timespec rem;                                                     \
    rem.tv_nsec = cpx_local_region_state;                                    \
                                                                             \
    cpx_local_rgnctx.exectx.stk_frame =                                      \
        (size_t)__builtin_frame_address(0) + sizeof(void*) *                 \
        (CPX_MACRO_SHARED_ARGS + 2);                                         \
                                                                             \
    cpx_intrinsic_get_tlsctx()->ctxset = 0;                                  \
    getcontext(&cpx_local_rgnctx.exectx.regs);                               \
                                                                             \
    if(!cpx_intrinsic_get_tlsctx()->ctxset)                                  \
    {                                                                        \
        cpx_local_rgnctx_ptr = &cpx_local_rgnctx;                            \
        cpx_exported_parallel((int)cpx_local_opt_arg, &cpx_local_rgnctx);    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_rgnctx_ptr = cpx_intrinsic_get_tlsctx()->rgnctx;           \
        __sync_fetch_and_sub(&cpx_local_rgnctx_ptr->head_rgn->busy, 1);      \
    }

#endif

/// barrier operator (default for parallel regions)
#define CPX_KEYWORD_BARRIER()                                                \
    }                                                                        \
    cpx_local_opt_arg =                                                      \
        __sync_fetch_and_sub(&cpx_local_rgnctx_ptr->head_rgn->bcount, 1);    \
                                                                             \
    if (!cpx_local_rgnctx_ptr->pix)                                          \
    {                                                                        \
        if (cpx_local_rgnctx_ptr->type == cpx_const_task)                    \
        {                                                                    \
            for (;;)                                                         \
            {                                                                \
                if (!cpx_exported_pick_nested_job(cpx_local_rgnctx_ptr))     \
                {                                                            \
                    if (cpx_local_rgnctx_ptr->bcount)                        \
                    {                                                        \
                        nanosleep(&req, &rem);                               \
                    }                                                        \
                    else                                                     \
                    {                                                        \
                        break;                                               \
                    }                                                        \
                }                                                            \
            }                                                                \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            while (cpx_local_rgnctx_ptr->bcount)                             \
            {                                                                \
               sem_wait(&cpx_local_rgnctx_ptr->head->wakesig);               \
            }                                                                \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
    }                                                                        \
    else                                                                     \
    {                                                                        \
        if (!cpx_local_opt_arg)                                              \
        {                                                                    \
            sem_post(&cpx_local_rgnctx_ptr->head->wakesig);                  \
        }                                                                    \
        cpx_local_rgnctx_ptr->curr->rgnctx = cpx_local_rgnctx_ptr->prev_nest;\
        longjmp(*cpx_local_rgnctx_ptr->idle, 1);                             \
    }

/// barrier skipping operator (exits parallel and serial regions)
#define CPX_KEYWORD_BREAK()                                                  \
    }                                                                        \
    if (cpx_local_region_state == CPX_CONST_STATE_PARALLEL)                  \
    {                                                                        \
        if (!cpx_local_rgnctx_ptr->pix)                                      \
        {                                                                    \
            if (cpx_local_rgnctx_ptr->type == cpx_const_task)                \
            {                                                                \
                for (;;)                                                     \
                {                                                            \
                    if (!cpx_exported_pick_nested_job(cpx_local_rgnctx_ptr)) \
                    {                                                        \
                        if (cpx_local_rgnctx_ptr->busy)                      \
                        {                                                    \
                            nanosleep(&req, &rem);                           \
                        }                                                    \
                        else                                                 \
                        {                                                    \
                            break;                                           \
                        }                                                    \
                    }                                                        \
                }                                                            \
            }                                                                \
            else                                                             \
            {                                                                \
                while (cpx_local_rgnctx_ptr->busy) nanosleep(&req, &rem);    \
            }                                                                \
            cpx_local_rgnctx_ptr->head->rgnctx =                             \
                cpx_local_rgnctx_ptr->prev_nest;                             \
        }                                                                    \
        else                                                                 \
        {                                                                    \
            longjmp(*cpx_local_rgnctx_ptr->idle, 1);                         \
        }                                                                    \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        (cpx_local_region_state == CPX_CONST_STATE_SERIAL_SYNC) ?            \
            cpx_exported_unsync(cpx_local_opt_arg) :                         \
            cpx_exported_unsync_default();                                   \
    }

/// serial operator
#define CPX_KEYWORD_SERIAL(...)                                              \
{                                                                            \
    int cpx_local_region_state = CPX_CONST_STATE_SERIAL_SYNC;                \
    cpx_datatype_region_context* cpx_local_rgnctx_ptr = 0;                   \
    size_t cpx_local_opt_arg = cpx_intrinsic_get_arg(1, ## __VA_ARGS__, 0);  \
                                                                             \
                                                                             \
    if (CPX_MACRO_ONE_PTR(__VA_ARGS__))                                      \
    {                                                                        \
        cpx_exported_sync(cpx_local_opt_arg);                                \
    }                                                                        \
    else                                                                     \
    {                                                                        \
        cpx_local_region_state = CPX_CONST_STATE_SERIAL_DEFAULT;             \
        cpx_exported_sync_default();                                         \
    }

/// serial operator with try-lock semantics

#define CPX_MACRO_TRY_serial(a) cpx_exported_trysync((size_t)a)
#define CPX_MACRO_GET_serial(a) cpx_intrinsic_get_arg(1, (size_t)a)

#define try(a)                                                               \
                                                                             \
    if (CPX_MACRO_TRY_##a)                                                   \
    {                                                                        \
        int cpx_local_region_state = CPX_CONST_STATE_SERIAL_SYNC;            \
        size_t cpx_local_opt_arg = CPX_MACRO_GET_##a;


/// ---------------------------------------------------------------------------
/// Language-Defined Data Types and Modifiers
/// ---------------------------------------------------------------------------

/// modifier to ensure shared variables are not kept in registers
#define CPX_KEYWORD_SHARED volatile

/// ---------------------------------------------------------------------------
/// Language-Defined (Built-In) Functions
/// ---------------------------------------------------------------------------

/// system-unique parallel task ID
#define CPX_KEYWORD_PID() cpx_exported_pid()
/// parallel region-unique task index
#define CPX_KEYWORD_PIX() cpx_exported_pix()
/// the number of tasks in the current parallel region
#define CPX_KEYWORD_PNO() cpx_exported_pno()

/// signaling operator to pass control to a task
#define CPX_KEYWORD_WAKE cpx_exported_wake
/// control suspension operator to wait until the control is returned
#define CPX_KEYWORD_WAIT(...) cpx_exported_wait                              \
                               (CPX_MACRO_ONE_ARG(__VA_ARGS__), ## __VA_ARGS__)

/// ---------------------------------------------------------------------------
/// Language-Defined Auxiliary Macros
/// ---------------------------------------------------------------------------

/// data-decomposition macros
#define CPX_MACRO_RANGE_START(size) cpx_exported_range_start(size)
#define CPX_MACRO_RANGE_END(size) cpx_exported_range_end(size)

/// ===========================================================================
/// C= End of Header
/// ===========================================================================

#ifdef __cplusplus
}
#endif

#endif ///_CPX_HEADER_H_
